/****************************************************************************
 *                                                                          *
 * Compare F0 contours (compf0.c)                                           *
 *                                                                          *
 * Implemented by Paul Bagshaw, Centre for Speech Technology Research,      *
 * University of Edinburgh, 80 South Bridge, Edinburgh EH1 1HN.             *
 *                                                                          *
 * 4th July 1991 (Modified 17/10/91)                                        *
 *                                                                          *
 ****************************************************************************/

#include <math.h>
#include <stdio.h>
#include <values.h>
#include "cleanup.h"

#define DEBUG 0

typedef enum {
  CANT_OPEN, CANT_READ, CANT_WRITE, INTRPOL, UX_EOF, MISUSE
  } error_types;

typedef enum {
  BOTH_VOICED, PDA_VOICED_ONLY, PDA_UNVOICED_ONLY, BOTH_UNVOICED
  } status_types;

typedef struct {
  unsigned int uv_errs, v_errs, doubles, halves, both_v, both_v_pge, both_uv;
  float uverr_duration, verr_duration;
  double sum_d, sum_dpge, sum_dpge_sqr;
} STATISTICS_;

static char program[20] = "compf0";

void main (argc, argv)
     int argc;
     char *argv[];
{

  void error (), initialise_statistics (), modify_statistics ();
  int read_contour ();
  float interpolate ();

  FILE *stdf0_file, *pdaf0_file, *stats_file = NULL;
  char i = '\0';
  STATISTICS_ stats;
  float stdtime, stdf0, pdatime, pdaf0;
  float prev_stdt = 0.0, prev_stdf0 = 0.0, prev_pdat = 0.0, prev_pdaf0 = 0.0;
  float new_stdf0 = 0.0, new_pdaf0 = 0.0, time, prev_time;
  float start_time = 0.0, end_time;
  status_types status = BOTH_UNVOICED, prev_status;

  /* parse parameters */
  if (argc != 3 && argc != 4) error (MISUSE, NULL);
  if ((stdf0_file = fopen (argv[1], "r")) == NULL) error (CANT_OPEN, argv[1]);
  if ((pdaf0_file = fopen (argv[2], "r")) == NULL) error (CANT_OPEN, argv[2]);
  if (argc == 4)
    if ((stats_file = fopen (argv[3], "a")) == NULL)
      error (CANT_OPEN, argv[3]);
  /* ignore Xmg fileheaders */
  while (!feof (stdf0_file) && i != '\14')
    if (!fread ((char *) &i, sizeof (char), 1, stdf0_file))
      error (CANT_READ, argv[1]);
  if (!fread ((char *) &i, sizeof (char), 1, stdf0_file))
    error (CANT_READ, argv[1]);
  if (i != '\n') error (CANT_READ, argv[1]);
  while (!feof (pdaf0_file) && i != '\14')
    if (!fread ((char *) &i, sizeof (char), 1, pdaf0_file))
      error (CANT_READ, argv[2]);
  if (!fread ((char *) &i, sizeof (char), 1, pdaf0_file))
    error (CANT_READ, argv[2]);
  if (i != '\n') error (CANT_READ, argv[2]);
  /* compare contours */
  initialise_statistics (&stats);
  read_contour (stdf0_file, &stdtime, &stdf0);
  read_contour (pdaf0_file, &pdatime, &pdaf0);
  time = (stdtime < pdatime) ? stdtime : pdatime;
  while (!feof (stdf0_file) || !feof (pdaf0_file)) {
    prev_time = time;
    if (stdtime < pdatime) {
      time = stdtime;
      new_stdf0 = stdf0;
      new_pdaf0 = interpolate (prev_pdat, prev_pdaf0, pdatime, pdaf0, time);
      prev_stdt = stdtime;
      prev_stdf0 = stdf0;
      read_contour (stdf0_file, &stdtime, &stdf0);
    }
    else if (pdatime < stdtime) {
      time = pdatime;
      new_stdf0 = interpolate (prev_stdt, prev_stdf0, stdtime, stdf0, time);
      new_pdaf0 = pdaf0;
      prev_pdat = pdatime;
      prev_pdaf0 = pdaf0;
      read_contour (pdaf0_file, &pdatime, &pdaf0);
    }
    else {
      time = stdtime = pdatime;
      new_stdf0 = stdf0;
      new_pdaf0 = pdaf0;
      prev_stdt = stdtime;
      prev_stdf0 = stdf0;
      read_contour (stdf0_file, &stdtime, &stdf0);
      prev_pdat = pdatime;
      prev_pdaf0 = pdaf0;
      read_contour (pdaf0_file, &pdatime, &pdaf0);
    }
    prev_status = status;
    if (new_stdf0 != 0.0 && new_pdaf0 != 0.0) status = BOTH_VOICED;
    else if (new_stdf0 == 0.0 && new_pdaf0 != 0.0) status = PDA_VOICED_ONLY;
    else if (new_stdf0 != 0.0 && new_pdaf0 == 0.0) status = PDA_UNVOICED_ONLY;
    else status = BOTH_UNVOICED;
    if (status != prev_status) {
      end_time = (status == BOTH_VOICED) ? time : prev_time;
      modify_statistics (prev_status, start_time, end_time, &stats);
      start_time = end_time;
    }
    if (status == BOTH_VOICED) {
      if ((new_stdf0 -  new_pdaf0) / new_stdf0 > 0.2)
	stats.halves++;
      else if ((new_pdaf0 -  new_stdf0) / new_stdf0 > 0.2)
	stats.doubles++;
      else {
	stats.sum_dpge += fabs (new_stdf0 - new_pdaf0);
	stats.sum_dpge_sqr += (new_stdf0 - new_pdaf0) *
	  (new_stdf0 - new_pdaf0);
	stats.both_v_pge++;
      }
      stats.sum_d += fabs (new_stdf0 - new_pdaf0);
      stats.both_v++;
    }
  }
  if (status == PDA_VOICED_ONLY || status == PDA_UNVOICED_ONLY) {
    end_time = (stdtime < pdatime) ? stdtime : pdatime;
    modify_statistics (status, start_time, end_time, &stats);
  }
  if (status == BOTH_VOICED) {
    if ((new_stdf0 -  new_pdaf0) / new_stdf0 > 0.2)
      stats.halves++;
    else if ((new_pdaf0 -  new_stdf0) / new_stdf0 > 0.2)
      stats.doubles++;
    else {
      stats.sum_dpge += fabs (new_stdf0 - new_pdaf0);
      stats.sum_dpge_sqr += (new_stdf0 - new_pdaf0) * (new_stdf0 - new_pdaf0);
      stats.both_v_pge++;
    }
    stats.sum_d += fabs (new_stdf0 - new_pdaf0);
    stats.both_v++;
    end_time = (stdtime < pdatime) ? stdtime : pdatime;
    modify_statistics (status, start_time, end_time, &stats);
    start_time = end_time;
    if (stdtime < pdatime) {
      status = PDA_VOICED_ONLY;
      end_time = pdatime;
    }
    else if (stdtime > pdatime) {
      status = PDA_UNVOICED_ONLY;
      end_time = stdtime;
    }
    if (status == PDA_VOICED_ONLY || status == PDA_UNVOICED_ONLY) {
      modify_statistics (status, start_time, end_time, &stats);
    }
  }
  /* display comparison statistics */
  if (stats_file == NULL) {
    printf ("Compare f0 contours: %s vs. %s\n", argv[1], argv[2]);
    printf (" Unvoiced regions classed as voiced by pda: %u ", stats.uv_errs);
    printf ("over %.4f ms\n", stats.uverr_duration);
    printf (" Voiced regions classed as unvoiced by pda: %u ", stats.v_errs);
    printf ("over %.4f ms\n", stats.verr_duration);
    printf (" Gross errors (doubles):\t%.2f %%\n", stats.doubles * 100.0 /
	    stats.both_v);
    printf ("               (halves):\t%.2f %%\n", stats.halves * 100.0 /
	    stats.both_v);
    printf (" Contour-deviation bar gross errors:\n    Average:\t\t%.2f Hz\n",
	    stats.sum_dpge / stats.both_v_pge);
    printf ("    Standard deviation:\t%.2f Hz\n",
	    sqrt ((stats.sum_dpge_sqr - stats.sum_dpge * stats.sum_dpge /
		   stats.both_v_pge) / (stats.both_v_pge - 1)));
  }
  else
    if (!fwrite ((STATISTICS_ *) &stats, sizeof (STATISTICS_), 1, stats_file))
      error (CANT_WRITE, argv[3]);
  exit (0);

}

void error (flag, name)
     error_types flag;
     char *name;
{

  fprintf (stderr, "%s: ", program);
  switch (flag) {
  case CANT_OPEN:
    fprintf (stderr, "cannot open file named %s", name);
    break;
  case CANT_READ:
    if (name == NULL)
      fprintf (stderr, "cannot read data from input file");
    else
      fprintf (stderr, "cannot read data from %s", name);
    break;
  case CANT_WRITE:
    fprintf (stderr, "cannot write to output file");
    break;
  case INTRPOL:
    fprintf (stderr, "strange input data - cannot interpolate");
    break;
  case UX_EOF:
    fprintf (stderr, "unexpected end of file for %s", name);
    break;
  case MISUSE:
    fprintf (stderr, "usage: %s standard_f0_file pda_f0_file", program);
    fprintf (stderr, " [appended_stats_file]");
    break;
  default:
    fprintf (stderr, "undefined error, %d occured", flag);
    break;
  }
  fprintf (stderr, "\n");
  exit (-1);

}

void initialise_statistics (p_stats)
     STATISTICS_ *p_stats;
{

  p_stats->uv_errs = 0;
  p_stats->v_errs = 0;
  p_stats->doubles = 0;
  p_stats->halves = 0;
  p_stats->both_v = 0;
  p_stats->both_v_pge = 0;
  p_stats->both_uv = 1;
  p_stats->uverr_duration = 0.0;
  p_stats->verr_duration = 0.0;
  p_stats->sum_d = 0.0;
  p_stats->sum_dpge = 0.0;
  p_stats->sum_dpge_sqr = 0.0;

}

int read_contour (f0_file, p_time, p_f0)
     FILE *f0_file;
     float *p_time, *p_f0;
{

  char i;
  long pos;

  if (feof (f0_file) || !fread ((char *) &i, sizeof (char), 1, f0_file) ||
      i == '\n') {
    *p_time = MAXFLOAT;
    *p_f0 = 0.0;
    return (0);
  }
  if (i == '=') {
    if (!fread ((char *) &i, sizeof (char), 1, f0_file) || i != '\n')
      error (CANT_READ, NULL);
    pos = ftell (f0_file);
    if (fscanf (f0_file, "%f ", p_time) != 1) error (CANT_READ, NULL);
    fseek (f0_file, pos, 0);
    *p_f0 = 0.0;
    return (0);
  }
  fseek (f0_file, (long) -sizeof (char), 1);
  if (fscanf (f0_file, "%f %f\n", p_time, p_f0) != 2) error (CANT_READ, NULL);
  return (1);

}

void modify_statistics (status, start_time, end_time, p_stats)
     status_types status;
     float start_time, end_time;
     STATISTICS_ *p_stats;
{

  switch (status) {
  case BOTH_VOICED:
    if (DEBUG)
      printf ("B V:\t%f - %f\n", start_time, end_time);
    break;
  case PDA_VOICED_ONLY:
    p_stats->uv_errs++;
    p_stats->uverr_duration += end_time - start_time;
    if (DEBUG)
      printf ("UV err:\t%f - %f\n", start_time, end_time);
    break;
  case PDA_UNVOICED_ONLY:
    p_stats->v_errs++;
    p_stats->verr_duration += end_time - start_time;
    if (DEBUG)
      printf ("V err:\t%f - %f\n", start_time, end_time);
    break;
  case BOTH_UNVOICED:
    p_stats->both_uv++;
    if (DEBUG)
      printf ("B UV:\t%f - %f\n", start_time, end_time);
    break;
  default:
    break;
  }

}

float interpolate (x1, y1, x2, y2, x)
     float x1, y1, x2, y2, x;
{

  float y, A, B;

  if (x < x1 || x > x2) {
    if (DEBUG)
      printf ("x1 = %f\ty1 = %f\nx2 = %f\ty2 = %f\nx = %f\n",
	      x1, y1, x2, y2, x);
    error (INTRPOL, NULL);
  }
  if (y1 == 0.0 || y2 == 0.0) return (0.0);
  A = (x2 - x) / (x2 - x1);
  B = 1 - A;
  y = A * y1 + B * y2;
  return (y);

}
